import logging
import pprint
import subprocess
from typing import Any, Callable, Dict, List, Tuple

import jinja2
import requests

from ..conf import GitlabCommon

LOG = logging.getLogger("mirror-tool")
SHARED_LABEL = "mirror-tool"

RunCmd = Callable[..., subprocess.CompletedProcess]


class GitlabException(RuntimeError):
    pass


class GitlabSession:
    def __init__(self, gitlab_info: GitlabCommon, run_cmd: RunCmd, dry_run: bool):
        for field in ("api_v4_url", "project_id", "push_url"):
            if not getattr(gitlab_info, field):
                raise GitlabException(
                    (
                        f"Cannot connect to GitLab: required config '{field}' is not set and "
                        "can't be determined automatically. Please run the tool within a GitLab pipeline "
                        "or set this field in the mirror-tool configuration file."
                    )
                )

        self.gitlab_info = gitlab_info
        self.api_v4_url = gitlab_info.api_v4_url
        self.project_id = gitlab_info.project_id
        self.requests = requests.Session()
        self.requests.headers["PRIVATE-TOKEN"] = gitlab_info.token_final

        self.run_cmd = run_cmd
        self.dry_run = dry_run

        jinja_loader = jinja2.DictLoader(
            {
                "title": self.gitlab_info.title,
                "description": self.gitlab_info.description,
                "comment.create": self.gitlab_info.comment.create,
                "comment.update": self.gitlab_info.comment.update,
            }
        )
        self.jinja_env = jinja2.Environment(loader=jinja_loader)
        self.jinja_args = {}

    def jinja_render(self, template_name, *args, **kwargs):
        kwargs.update(self.jinja_args)
        return self.jinja_env.get_template(template_name).render(*args, **kwargs)

    @property
    def project_mrs_url(self):
        return "".join(
            [self.api_v4_url, "/projects/", str(self.project_id), "/merge_requests"]
        )

    def response_ok(self, doing_what, response):
        if not response.ok:
            LOG.warning(
                "Unexpected response from GitLab: %s %s",
                response.status_code,
                response.reason,
            )
            LOG.warning("Response body: %s", response.json())
            raise GitlabException(f"Failed to {doing_what}")

    def run_git_silent(self, cmd, doing_what):
        try:
            self.run_cmd(cmd, silent=True)
        except Exception:
            # The unusual error handling here is because the push_url is
            # likely to contain a token which shouldn't be leaked, and the exception
            # is going to contain the command by default. So we drop that and raise
            # our own more vague exception.
            LOG.debug("Command failed", exc_info=True)
            raise GitlabException(f"Could not {doing_what}.")

    def ensure_pushed_to(self, revision, dest):
        if self.dry_run:
            LOG.info(
                "DRY RUN: if not using --dry-run, would now push %s to %s",
                revision,
                dest,
            )
            return

        self.run_git_silent(
            [
                "git",
                "push",
                self.gitlab_info.push_url_final,
                f"+{revision}:refs/heads/{dest}",
            ],
            f"push {revision} to {dest} branch",
        )

    def mutable_mr_attributes(self) -> Dict[str, Any]:
        return dict(
            title=self.jinja_render("title"),
            allow_collaboration=True,
            squash=False,
            labels=[SHARED_LABEL] + self.gitlab_info.labels,
            description=self.jinja_render("description"),
        )

    def create_mr_with_branches(self, src, dest) -> bool:
        """Make a merge request in the configured repo from 'src' to 'dest' branch.

        Returns True if MR created, False if MR not created due to conflict, or
        raises on error.
        """

        # https://docs.gitlab.com/ee/api/merge_requests.html#create-mr

        create_attrs = self.mutable_mr_attributes()
        create_attrs["source_branch"] = src
        create_attrs["target_branch"] = dest

        if self.dry_run:
            LOG.info(
                "DRY RUN: if not using --dry-run, would now create MR from %s to %s",
                src,
                dest,
            )
            LOG.info("Merge request details:\n%s", pprint.pformat(create_attrs))
            return True

        LOG.info("Creating GitLab merge request ...")

        response = self.requests.post(
            self.project_mrs_url,
            json=create_attrs,
        )

        LOG.info("GitLab response: %s", response.status_code)

        if response.status_code == 409:
            # This code is tolerated as meaning that an MR already exists.
            return False

        self.response_ok("create merge request", response)

        body = response.json()
        LOG.info("Created: %s", body.get("web_url") or "<unknown merge request URL>")
        self.add_comment(body, self.jinja_render("comment.create"))
        return True

    def update_mr(self, mr) -> None:
        url = "".join([self.project_mrs_url, "/", str(mr["iid"])])

        update_attrs = self.mutable_mr_attributes()

        if self.dry_run:
            LOG.info("DRY RUN: if not using --dry-run, would now update MR %s.", url)
            LOG.info("Merge request details:\n%s", pprint.pformat(update_attrs))
            return

        response = self.requests.put(url, json=update_attrs)

        self.response_ok("update merge request", response)

        self.add_comment(mr, self.jinja_render("comment.update"))

        LOG.info("Updated: %s", mr.get("web_url") or "<unknown merge request URL>")

    def add_comment(self, mr, body):
        if not body:
            return

        # https://docs.gitlab.com/ee/api/notes.html#create-new-merge-request-note
        iid = mr["iid"]
        url = "".join([self.project_mrs_url, "/", str(iid), "/notes"])

        if self.dry_run:
            LOG.info("DRY RUN: if not using --dry-run, would now add comment to MR.")
            LOG.info("Comment body:\n%s", body)
            return

        response = self.requests.post(url, json=dict(body=body))
        self.response_ok("add comment", response)

        LOG.info("Commented on: %s", mr.get("web_url") or "<unknown merge request URL>")

    def find_mrs_with_fields(self, fields) -> Tuple[dict, dict]:
        """Searches for MRs matching passed fields.

        Returns a tuple of:

          (list of matching MRs created by mirror-tool,
           list of all matching MRs)
        """

        # Note: for the time being, lack of pagination here should be OK.
        #
        # - default sort will return newest MRs first.
        # - default pagination is to return first page only.
        #
        # So, default behavior is to return ~20 most recent MRs only.
        # It should be OK for the purpose of this tool.
        response = self.requests.get(
            self.project_mrs_url,
            params=fields,
        )

        self.response_ok("find merge request", response)

        mrs = response.json()
        out = []
        for mr in mrs:
            if SHARED_LABEL in (mr.get("labels") or []):
                out.append(mr)

        return (out, mrs)

    def find_single_mr(self, fields) -> dict:
        (mrs_own, mrs_all) = self.find_mrs_with_fields(fields)

        if mrs_own:
            mr = mrs_own[0]
            mr_url = mr.get("web_url") or "<unknown url>"
            LOG.info(
                "Found existing merge request: %s",
                mr_url,
            )
            return mr

        if mrs_all:
            mr_url = mrs_all[0].get("web_url") or "<unknown url>"

            # If we get here, that means we found an MR for the right
            # branches but it apparently wasn't created by us?
            # That makes it not safe to proceed.
            raise GitlabException(
                f"An existing merge request {mr_url} was found, but it was apparently "
                "not created by mirror-tool! Refusing to update. Please close the MR manually "
                "(if it is safe to do so), then re-run the tool."
            )

        # If we get here, that means we didn't find any MR at all which
        # is unexpected.
        raise GitlabException(
            "Failed to create MR due to conflict, "
            "but also failed to locate an existing MR!"
        )

    def create_or_update_mr(self, create_fn, update_fn, find_fields):
        if create_fn():
            return

        LOG.info("GitLab merge request seems to already exist, searching...")
        mr = self.find_single_mr(find_fields)
        update_fn(mr)

    def revision_in_remote_branch(self, revision: str, branch: str) -> bool:
        """Returns True if 'revision' appears to be reachable from remote 'branch'."""

        # Make sure we have latest version of that branch in a local ref...
        self.run_git_silent(
            [
                "git",
                "fetch",
                self.gitlab_info.push_url_final,
                f"+refs/heads/{branch}:refs/mirror-tool/dest-branch",
            ],
            f"fetch remote branch {branch}",
        )

        # Is that revision already reachable from target branch?
        # Note: this will also fail if 'revision' isn't even recognizable as a revision.
        # But in that case we still know it's not in the dest branch, so returning
        # False is reasonable.
        proc = self.run_cmd(
            [
                "git",
                "merge-base",
                "--is-ancestor",
                revision,
                "refs/mirror-tool/dest-branch",
            ],
            check=False,
        )
        if proc.returncode == 0:
            LOG.info(
                "Revision %s is already reachable from remote %s.",
                revision,
                branch,
            )
            return True

        return False



import logging
import os
import subprocess
from dataclasses import dataclass, field
from datetime import datetime
from typing import Generator, List

from .shared import Mirror

LOG = logging.getLogger("mirror-tool")
COMMIT_LIMIT = int(os.getenv("MIRROR_TOOL_COMMIT_LIMIT") or "20")


class GitParseError(RuntimeError):
    pass


@dataclass
class Commit:
    """Contains info on a single commit."""

    revision: str
    """The commit's revision (SHA1)."""

    revision_abbrev: str
    """The commit's abbreviated revision."""

    author_name: str
    """The author's name."""

    author_email: str
    """The author's email."""

    author_email_local: str
    """The local part of the author's email (prior to the '@' character)."""

    author_datetime: datetime
    """The 'author' timestamp."""

    committer_name: str
    """The committer's name."""

    committer_email: str
    """The committer's email."""

    committer_email_local: str
    """The local part of the committer's email (prior to the '@' character)."""

    committer_datetime: datetime
    """The 'committer' timestamp."""

    subject: str
    """The subject line from the commit message."""

    body: str = ""
    """The body of the commit message."""

    url: str = ""
    """A URL at which this commit can possibly be accessed.

    This value is guessed based on the Git hosting service which appears to be in use
    for the mirrored repo. It might be wrong or missing in some cases.
    """

    @classmethod
    def from_log(cls, log: bytes) -> Generator["Commit", None, None]:
        # Parse a git log into Commit objects.
        # Log should be generated by command:
        # git log --pretty=format:%H%n%h%n%an%n%ae%n%al%n%at%n%cn%n%ce%n%cl%n%ct%n%s%n%b -z

        commit_logs = log.split(b"\x00")

        for commit_log in commit_logs:
            lines = commit_log.decode("utf-8").splitlines()
            if not lines:
                continue

            kwargs = {}

            kwargs["revision"] = lines.pop(0)
            kwargs["revision_abbrev"] = lines.pop(0)
            kwargs["author_name"] = lines.pop(0)
            kwargs["author_email"] = lines.pop(0)
            kwargs["author_email_local"] = lines.pop(0)
            kwargs["author_datetime"] = datetime.utcfromtimestamp(int(lines.pop(0)))
            kwargs["committer_name"] = lines.pop(0)
            kwargs["committer_email"] = lines.pop(0)
            kwargs["committer_email_local"] = lines.pop(0)
            kwargs["committer_datetime"] = datetime.utcfromtimestamp(int(lines.pop(0)))
            kwargs["subject"] = lines.pop(0)
            kwargs["body"] = "\n".join(lines)

            yield Commit(**kwargs)


@dataclass
class UpdateInfo:
    """Contains info on an update for a single mirror."""

    mirror: Mirror
    """The mirror associated with this update."""

    commits: List[Commit] = field(default_factory=list)
    """The commits pulled in by the update. The first entry is the tip of the update."""

    commit_count: int = 0
    """The number of commit(s) being pulled by the update."""

    commit_elided_count: int = 0
    """The number of commit(s) being pulled in, but omitted from UpdateInfo because
    there were too many commits.

    Equal to commit_count - len(commits).
    """

    changed: bool = False
    """True if there were any changes at all."""


def add_urls(mirror: Mirror, commits: List[Commit]):
    if not mirror.url.startswith("https://"):
        return

    if "github" in mirror.url:
        for commit in commits:
            commit.url = f"{mirror.url}/commit/{commit.revision}"
    elif "gitlab" in mirror.url:
        for commit in commits:
            commit.url = f"{mirror.url}/-/commit/{commit.revision}"


def get_update_info(
    rev_to: str,
    mirror: Mirror,
    rev_from: str = "HEAD",
    commit_limit: int = COMMIT_LIMIT,
) -> UpdateInfo:
    logs = subprocess.check_output(
        [
            "git",
            "log",
            "-z",
            "--pretty=format:%H%n%h%n%an%n%ae%n%al%n%at%n%cn%n%ce%n%cl%n%ct%n%s%n%b",
            f"{rev_from}..{rev_to}",
        ],
        text=False,
    )

    commits = list(Commit.from_log(logs))
    add_urls(mirror, commits)
    count = len(commits)
    elided = 0

    if len(commits) > commit_limit:
        commits = commits[:commit_limit]
        elided = count - len(commits)

    changed = True if count else False

    return UpdateInfo(
        mirror=mirror,
        changed=changed,
        commit_count=count,
        commit_elided_count=elided,
        commits=commits,
    )



import datetime
import os
from typing import Any

import jinja2


def jinja_args(**kwargs) -> dict[str, Any]:
    now = datetime.datetime.utcnow()
    out = dict(
        env=os.environ,
        datetime_iso8601=now.strftime("%Y-%m-%dT%H:%M:%SZ"),
        datetime_minute=now.strftime("%Y-%m-%d %H:%M"),
        datetime_day=now.strftime("%Y-%m-%d"),
        datetime_week=now.strftime("%Ywk%U"),
    )
    out.update(kwargs)
    return out


def jinja_validate(template: str, **kwargs):
    loader = jinja2.DictLoader({"template": template})
    env = jinja2.Environment(undefined=jinja2.StrictUndefined)
    loader.load(env, "template").render(jinja_args(**kwargs))



import os
from typing import Any, Dict


def environ_with_git_config(
    git_config: Dict[str, Any], input_environ: Dict[str, str] = os.environ
) -> Dict[str, str]:
    out = input_environ.copy()
    count = int(input_environ.get("GIT_CONFIG_COUNT") or "0")

    for key, val in git_config.items():
        out[f"GIT_CONFIG_KEY_{count}"] = key
        out[f"GIT_CONFIG_VALUE_{count}"] = str(val)
        count += 1

    out["GIT_CONFIG_COUNT"] = str(count)
    return out



import datetime
import os
from dataclasses import asdict, dataclass, field
from typing import Any, Dict, List

import jsonschema
from ruamel.yaml import YAML

from .git_info import Commit, UpdateInfo
from .jinja import jinja_validate
from .shared import Mirror

CONFIG_SCHEMA = {
    "$schema": "https://json-schema.org/draft/2020-12/schema",
    "definitions": {
        "jinjaTemplateLarge": {"type": "string", "minLength": 0, "maxLength": 8000},
        "relativeDir": {
            "allOf": [
                {
                    "type": "string",
                    "minLength": 1,
                    "maxLength": 4000,
                },
                {"not": {"type": "string", "pattern": r"\.\."}},
                {"not": {"type": "string", "pattern": r"//"}},
                {"not": {"type": "string", "pattern": r"^/"}},
            ],
        },
        "mirror": {
            "type": "object",
            "properties": {
                "url": {
                    "type": "string",
                    "minLength": 2,
                    "maxLength": 4000,
                },
                "ref": {
                    "type": "string",
                    "minLength": 2,
                    "maxLength": 4000,
                },
                "dir": {"$ref": "#/definitions/relativeDir"},
            },
            "required": ["url", "ref"],
            "additionalProperties": False,
        },
        "mirrorList": {
            "type": "array",
            "items": {"$ref": "#/definitions/mirror"},
        },
        "gitlabMerge": {
            "type": "object",
            "properties": {
                "enabled": {"type": "boolean"},
                "src": {"type": "string", "minLength": 1, "maxLength": 200},
                "dest": {"type": "string", "minLength": 1, "maxLength": 200},
                "title": {"type": "string", "minLength": 1, "maxLength": 2000},
                "description": {"type": "string", "minLength": 0, "maxLength": 8000},
                "token": {"type": "string", "pattern": r"^\$.+"},
                "labels": {
                    "type": "array",
                    "items": {"type": "string", "minLength": 1, "maxLength": 50},
                },
                "comment": {
                    "type": "object",
                    "properties": {
                        "create": {"$ref": "#/definitions/jinjaTemplateLarge"},
                        "update": {"$ref": "#/definitions/jinjaTemplateLarge"},
                    },
                    "additionalProperties": False,
                },
            },
            "additionalProperties": False,
        },
    },
    "type": "object",
    "properties": {
        "mirror": {"$ref": "#/definitions/mirrorList"},
        "gitlab_merge": {"$ref": "#/definitions/gitlabMerge"},
        "gitlab_promote": {
            "type": "array",
            # TODO: actually it's not 100% identical to gitlabMerge.
            "items": {"$ref": "#/definitions/gitlabMerge"},
        },
        "git_config": {"type": "object"},
        "commitmsg": {
            "type": "string",
            "minLength": 1,
            "maxLength": 8000,
        },
    },
    "required": [],
    "additionalProperties": False,
}


@dataclass
class GitlabMergeComments:
    create: str = ""
    update: str = ""


@dataclass
class GitlabCommon:
    src: str = "latest"
    dest: str = "main"
    title: str = "Update mirror"

    # This MUST be '$ENV_VAR_NAME'
    token: str = "$GITLAB_MIRROR_TOKEN"

    labels: List[str] = field(default_factory=list)
    description: str = "Automated update of dependencies."
    comment: GitlabMergeComments = field(default_factory=GitlabMergeComments)

    # The defaults here assume that we are running from a gitlab CI pipeline.
    # Predefined vars are documented at:
    # https://docs.gitlab.com/ee/ci/variables/predefined_variables.html
    api_v4_url: str = field(default_factory=lambda: os.environ.get("CI_API_V4_URL"))
    project_id: int = field(
        default_factory=lambda: int(os.environ.get("CI_PROJECT_ID") or "0")
    )

    push_url: str = field(
        default_factory=lambda: os.environ.get("CI_PROJECT_URL") or ""
    )

    @property
    def token_final(self) -> str:
        if not self.token.startswith("$"):
            raise ValueError(
                f"Invalid token: '{self.token}'. Must be of form '$ENV_VAR_NAME'."
            )
        key = self.token[1:]
        if key not in os.environ:
            raise ValueError(f"GitLab token not available in {self.token}.")

        return os.environ[key]

    @property
    def push_url_final(self) -> str:
        return self._token_auth(self.push_url)

    def _token_auth(self, url: str):
        if not url.startswith("https://"):
            # Not using token-based auth
            return url
        if "@" in url:
            # Maybe the caller already set up auth in the URL
            return url

        url_noscheme = url[len("https://") :]
        return "".join(["https://", "token:", self.token_final, "@", url_noscheme])


@dataclass
class GitlabMerge(GitlabCommon):
    enabled: bool = False


@dataclass
class GitlabPromote(GitlabCommon):
    title: str = "Promote changes"
    description: str = "Automated promotion between branches."

    @property
    def working_branch(self) -> str:
        return f"mirror-tool/promote-{self.src}-to-{self.dest}"


# Objects used for jinja validation purposes.
VALIDATE_COMMIT = Commit(
    revision="55810cd62082f26ec39a9df332af1aa9db6e6b91",
    revision_abbrev="55810cd",
    author_name="Rohan McGovern",
    author_email="rohan@mcgovern.id.au",
    author_email_local="rohan",
    author_datetime=datetime.datetime(2022, 6, 6, 22, 33, 34),
    committer_name="GitHub",
    committer_email="noreply@github.com",
    committer_email_local="noreply",
    committer_datetime=datetime.datetime(2022, 6, 6, 22, 33, 34),
    subject="Merge pull request #33 from rohanpm/empty-log",
    body="Fix a crash when commit log has an empty entry",
)

VALIDATE_MIRROR = Mirror(
    url="https://example.com/foo",
    ref="refs/heads/quux",
)

VALIDATE_UPDATEINFO = UpdateInfo(
    mirror=VALIDATE_MIRROR, commits=[VALIDATE_COMMIT], commit_count=1, changed=True
)

# This is straight from https://docs.gitlab.com/ee/api/merge_requests.html#get-single-mr
VALIDATE_MERGEREQUEST = {
    "id": 155016530,
    "iid": 133,
    "project_id": 15513260,
    "title": "Manual job rules",
    "description": "",
    "state": "opened",
    "created_at": "2022-05-13T07:26:38.402Z",
    "updated_at": "2022-05-14T03:38:31.354Z",
    "merge_user": None,
    "merged_at": None,
    "closed_by": None,
    "closed_at": None,
    "target_branch": "master",
    "source_branch": "manual-job-rules",
    "user_notes_count": 0,
    "upvotes": 0,
    "downvotes": 0,
    "author": {
        "id": 4155490,
        "username": "marcel.amirault",
        "name": "Marcel Amirault",
        "state": "active",
        "avatar_url": "https://gitlab.com/uploads/-/system/user/avatar/4155490/avatar.png",
        "web_url": "https://gitlab.com/marcel.amirault",
    },
    "assignees": [],
    "assignee": None,
    "reviewers": [],
    "source_project_id": 15513260,
    "target_project_id": 15513260,
    "labels": [],
    "draft": False,
    "work_in_progress": False,
    "milestone": None,
    "merge_when_pipeline_succeeds": False,
    "merge_status": "can_be_merged",
    "sha": "e82eb4a098e32c796079ca3915e07487fc4db24c",
    "merge_commit_sha": None,
    "squash_commit_sha": None,
    "discussion_locked": None,
    "should_remove_source_branch": None,
    "force_remove_source_branch": True,
    "reference": "!133",
    "references": {
        "short": "!133",
        "relative": "!133",
        "full": "marcel.amirault/test-project!133",
    },
    "web_url": "https://gitlab.com/marcel.amirault/test-project/-/merge_requests/133",
    "time_stats": {
        "time_estimate": 0,
        "total_time_spent": 0,
        "human_time_estimate": None,
        "human_total_time_spent": None,
    },
    "squash": False,
    "task_completion_status": {"count": 0, "completed_count": 0},
    "has_conflicts": False,
    "blocking_discussions_resolved": True,
    "approvals_before_merge": None,
    "subscribed": True,
    "changes_count": "1",
    "latest_build_started_at": "2022-05-13T09:46:50.032Z",
    "latest_build_finished_at": None,
    "first_deployed_to_production_at": None,
    "head_pipeline": {
        "id": 538317940,
        "iid": 1877,
        "project_id": 15513260,
        "sha": "1604b0c46c395822e4e9478777f8e54ac99fe5b9",
        "ref": "refs/merge-requests/133/merge",
        "status": "failed",
        "source": "merge_request_event",
        "created_at": "2022-05-13T09:46:39.560Z",
        "updated_at": "2022-05-13T09:47:20.706Z",
        "web_url": "https://gitlab.com/marcel.amirault/test-project/-/pipelines/538317940",
        "before_sha": "1604b0c46c395822e4e9478777f8e54ac99fe5b9",
        "tag": False,
        "yaml_errors": None,
        "user": {
            "id": 4155490,
            "username": "marcel.amirault",
            "name": "Marcel Amirault",
            "state": "active",
            "avatar_url": "https://gitlab.com/uploads/-/system/user/avatar/4155490/avatar.png",
            "web_url": "https://gitlab.com/marcel.amirault",
        },
        "started_at": "2022-05-13T09:46:50.032Z",
        "finished_at": "2022-05-13T09:47:20.697Z",
        "committed_at": None,
        "duration": 30,
        "queued_duration": 10,
        "coverage": None,
        "detailed_status": {
            "icon": "status_failed",
            "text": "failed",
            "label": "failed",
            "group": "failed",
            "tooltip": "failed",
            "has_details": True,
            "details_path": "/marcel.amirault/test-project/-/pipelines/538317940",
            "illustration": None,
            "favicon": "/assets/ci_favicons/favicon_status_failed-41304d7f7e3828808b0c26771f0309e55296819a9beea3ea9fbf6689d9857c12.png",
        },
    },
    "diff_refs": {
        "base_sha": "1162f719d711319a2efb2a35566f3bfdadee8bab",
        "head_sha": "e82eb4a098e32c796079ca3915e07487fc4db24c",
        "start_sha": "1162f719d711319a2efb2a35566f3bfdadee8bab",
    },
    "merge_error": None,
    "first_contribution": False,
    "user": {"can_merge": True},
}


class Config:
    def __init__(self, raw):
        self._raw = raw

    @property
    def gitlab_merge(self) -> GitlabMerge:
        raw = (self._raw.get("gitlab_merge") or {}).copy()
        raw_comment = raw.get("comment") or {}
        raw["comment"] = GitlabMergeComments(**raw_comment)
        return GitlabMerge(**raw)

    @property
    def gitlab_promote(self) -> List[GitlabPromote]:
        out = []
        raw = self._raw.get("gitlab_promote") or []
        for elem in raw:
            elem = elem.copy()
            raw_comment = elem.get("comment") or {}
            elem["comment"] = GitlabMergeComments(**raw_comment)
            out.append(GitlabPromote(**elem))
        return out

    @property
    def git_config(self) -> Dict[str, Any]:
        return self._raw.get("git_config") or {}

    @property
    def mirrors(self) -> List[Mirror]:
        out = []
        for elem in self._raw.get("mirror") or []:
            out.append(Mirror(**elem))
        return out

    @property
    def commitmsg(self) -> str:
        return (
            self._raw.get("commitmsg")
            or "merging {{mirror.dir}} at {{datetime_minute}}"
        )

    def validate(self) -> None:
        jsonschema.validate(self._raw, CONFIG_SCHEMA)

        mirror_dirs = {mirror.dir for mirror in self.mirrors}
        if len(mirror_dirs) != len(self.mirrors):
            raise jsonschema.ValidationError(
                message="Multiple mirrors defined using same dir",
                path=["mirror"],
                instance=self.mirrors,
            )

        self.jinja_validate_all()

    def jinja_validate_all(self):
        updates = [VALIDATE_UPDATEINFO, VALIDATE_UPDATEINFO]
        jinja_templates = []
        jinja_templates.append(
            (["commitmsg"], self.commitmsg, asdict(VALIDATE_UPDATEINFO))
        )

        def append_gitlab_common(base_path, instance, **kwargs):
            jinja_templates.append(
                (base_path + ["title"], getattr(instance, "title"), kwargs)
            )
            jinja_templates.append(
                (base_path + ["description"], getattr(instance, "description"), kwargs)
            )

            comment = getattr(instance, "comment")
            jinja_templates.append(
                (base_path + ["comment", "create"], comment.create, kwargs)
            )
            jinja_templates.append(
                (base_path + ["comment", "update"], comment.update, kwargs)
            )

        append_gitlab_common(["gitlab_merge"], self.gitlab_merge, updates=updates)

        for i, elem in enumerate(self.gitlab_promote):
            base_path = ["gitlab_promote", i]
            append_gitlab_common(base_path, elem, src_mr=VALIDATE_MERGEREQUEST)

        for path, template, kwargs in jinja_templates:
            self.jinja_validate(template, path, **kwargs)

    def jinja_validate(self, template, path, **kwargs):
        try:
            jinja_validate(template, **kwargs)
        except Exception as exc:
            raise jsonschema.ValidationError(
                message="Invalid Jinja template: %s" % str(exc),
                path=path,
                instance="<jinja template starting: %r...>" % (template[:50]),
                cause=exc,
            )

    @classmethod
    def from_file(cls, filename=".mirror-tool.yaml") -> "Config":
        with open(filename, "rt") as f:
            raw = YAML(typ="safe").load(f)
        return cls(raw)



import textwrap

import jinja2

from ..conf import Config

TEMPLATE = textwrap.dedent(
    """
    # This config was generated by 'mirror-tool gitlab-ci-yml'.
    # Use 'include' to load it from your main .gitlab-ci.yml.
    .mirror-tool-image:
      image:
        name: {{ image }}
        entrypoint: [""]
      tags: [docker]

    .mirror-tool-deploy:
      extends: .mirror-tool-image
      stage: {{ deploy_stage }}
      resource_group: mirror-tool

    "mirror-tool: validate config":
      extends: .mirror-tool-image
      script: mirror-tool validate-config
      rules:
      - if: '$CI_PIPELINE_SOURCE == "merge_request_event"'
      - if: '$CI_PIPELINE_SOURCE == "push"'

    "mirror-tool: update":
      extends: .mirror-tool-deploy
      script: mirror-tool update
      rules:
      - if: '{{ token_var }} && $MANUAL_UPDATE == "1" && $CI_PIPELINE_SOURCE == "web" && $CI_COMMIT_BRANCH == "{{ update_branch }}"'
      - if: '{{ token_var }} && $CI_PIPELINE_SOURCE == "schedule" && $CI_COMMIT_BRANCH == "{{ update_branch }}"'

    {% if promote_src_branches %}
    "mirror-tool: promote":
      extends: .mirror-tool-deploy
      script: mirror-tool promote
      rules:
      - if: '{{ token_var }} && $MANUAL_PROMOTE == "1" && $CI_PIPELINE_SOURCE == "web"'
    {%- for branch in promote_src_branches %}
      - if: '{{ token_var }} && $CI_PIPELINE_SOURCE == "push" && $CI_COMMIT_BRANCH == "{{ branch }}"'
    {%- endfor %}
    {% endif %}
"""
)


def render_ci_template_from_args(**kwargs) -> str:
    env = jinja2.Environment()
    loader = jinja2.DictLoader({"template": TEMPLATE})
    return loader.load(env, "template").render(**kwargs)


def render_ci_template_from_config(conf: Config) -> str:
    kwargs = {
        # Not everything in here is actually configurable right now.
        # Some values extracted as args are things which I think might need to
        # become configurable later on.
        "image": "quay.io/rmcgover/mirror-tool:latest",
        "deploy_stage": "deploy",
        "token_var": conf.gitlab_merge.token,
        "update_branch": conf.gitlab_merge.dest,
        "promote_src_branches": [p.src for p in conf.gitlab_promote],
    }
    return render_ci_template_from_args(**kwargs)



import logging

from ..conf import GitlabPromote
from ..jinja import jinja_args
from .common import SHARED_LABEL, GitlabSession, RunCmd

LOG = logging.getLogger("mirror-tool")


class GitlabPromoteSession(GitlabSession):
    def __init__(
        self, gitlab_promote: GitlabPromote, run_cmd: RunCmd, dry_run: bool = False
    ):
        super().__init__(gitlab_promote, run_cmd, dry_run)
        self.gitlab_promote = gitlab_promote

        self.jinja_args = jinja_args(updates=[])

    def ensure_pushed_to_workbranch(self, revision):
        return self.ensure_pushed_to(revision, self.gitlab_promote.working_branch)

    def create_mr(self) -> bool:
        return self.create_mr_with_branches(
            self.gitlab_promote.working_branch, self.gitlab_promote.dest
        )

    def find_merged_mr(self) -> dict:
        LOG.info("Looking for previous MRs to %s...", self.gitlab_promote.src)
        (mrs_own, _) = self.find_mrs_with_fields(
            {
                "target_branch": self.gitlab_promote.src,
                "labels": SHARED_LABEL,
                "state": "merged",
            },
        )

        if mrs_own:
            mr = mrs_own[0]
            mr_url = mr.get("web_url") or "<unknown url>"
            LOG.info(
                "Found existing merge request: %s",
                mr_url,
            )
            return mr

    def ensure_promotion_merge_request_exists(self):
        # Locate the latest submitted MR by ourselves to the target branch.
        src_mr = self.find_merged_mr()

        if not src_mr:
            LOG.info("No eligible MR for promotion.")
            return

        # Make this object available to Jinja contexts.
        self.jinja_args["src_mr"] = src_mr

        revision = src_mr["merge_commit_sha"]
        LOG.debug("Should promote: %s", revision)

        if self.revision_in_remote_branch(revision, self.gitlab_promote.dest):
            return

        self.ensure_pushed_to_workbranch(revision)

        self.create_or_update_mr(
            create_fn=self.create_mr,
            update_fn=self.update_mr,
            find_fields={
                "state": "opened",
                "source_branch": self.gitlab_promote.working_branch,
                "target_branch": self.gitlab_promote.dest,
            },
        )



import logging

from ..conf import GitlabMerge
from ..git_info import UpdateInfo
from ..jinja import jinja_args
from .common import GitlabSession

LOG = logging.getLogger("mirror-tool")


class GitlabUpdateSession(GitlabSession):
    def __init__(
        self,
        gitlab_merge: GitlabMerge,
        run_cmd,
        updates: list[UpdateInfo],
        dry_run: bool = False,
    ):
        super().__init__(gitlab_merge, run_cmd, dry_run)
        self.gitlab_merge = gitlab_merge
        self.updates = updates
        self.jinja_args = jinja_args(updates=self.updates)

    def ensure_pushed_to_src(self, revision):
        return self.ensure_pushed_to(revision, self.gitlab_merge.src)

    def create_mr(self) -> bool:
        return self.create_mr_with_branches(
            self.gitlab_merge.src, self.gitlab_merge.dest
        )

    def is_mr_uptodate(self, mr, revision) -> bool:
        web_url = mr.get("web_url") or "<unknown url>"
        LOG.info("Checking existing MR %s ...", web_url)

        mr_revision = mr["sha"]

        # Make sure we have the MR revision.
        self.run_git_silent(
            [
                "git",
                "fetch",
                self.gitlab_info.push_url_final,
                f"+{mr_revision}:refs/mirror-tool/existing-mr",
            ],
            f"fetch remote revision {mr_revision}",
        )

        if (
            self.run_cmd(
                ["git", "diff", "--quiet", revision, mr_revision], check=False
            ).returncode
            != 0
        ):
            # There are differences in content => it's not up-to-date
            LOG.info(
                "MR %s needs an update: there are differences in content.", web_url
            )
            return False

        # TODO: there could still be differences in the mutable MR attributes.
        # What to do about it?
        # - If we insist on updating MRs for any difference, it can create a lot of noise.
        # - If we don't update MRs, then config changes don't take effect until the next
        #   time an MR is created with different content.
        # - Does it need to be configurable?
        LOG.info("MR %s does not need an update.", web_url)
        return True

    def ensure_merge_request_exists(self, revision="HEAD"):
        if self.revision_in_remote_branch(revision, self.gitlab_merge.dest):
            return

        # Let's see if there's already an MR by us between src and dest branch.
        find_fields = {
            "state": "opened",
            "source_branch": self.gitlab_merge.src,
            "target_branch": self.gitlab_merge.dest,
        }
        (ours, _) = self.find_mrs_with_fields(find_fields)
        if ours and self.is_mr_uptodate(ours[0], revision):
            # Don't need to do anything.
            return

        # First have to make sure it's pushed.
        self.ensure_pushed_to_src(revision)

        self.create_or_update_mr(
            create_fn=self.create_mr,
            update_fn=self.update_mr,
            find_fields=find_fields,
        )



#!/usr/bin/env python3
import argparse
import logging
import os
import subprocess
import sys
from dataclasses import asdict
from typing import Optional

import jinja2
from jsonschema.exceptions import ValidationError

from .conf import Config, Mirror
from .git_config import environ_with_git_config
from .git_info import UpdateInfo, get_update_info
from .gitlab import (
    GitlabPromoteSession,
    GitlabUpdateSession,
    render_ci_template_from_config,
)
from .jinja import jinja_args

LOG = logging.getLogger("mirror-tool")


def add_dryrun(parser: argparse.ArgumentParser):
    parser.add_argument(
        "--dry-run",
        default=False,
        action="store_true",
        help="Avoid making changes to remote systems, log what would be done",
    )


class MirrorTool:
    def __init__(self):
        self.args: Optional[argparse.Namespace] = None
        self._config: Optional[Config] = None

    @property
    def parser(self) -> argparse.ArgumentParser:
        parser = argparse.ArgumentParser()
        parser.set_defaults(func=self.no_command)

        parser.add_argument(
            "--conf",
            "-c",
            type=str,
            default=".mirror-tool.yaml",
            help="Path to configuration file for mirror-tool",
        )
        subparsers = parser.add_subparsers()

        validate_config = subparsers.add_parser(
            "validate-config", help="Validate a mirror-tool configuration file"
        )
        validate_config.set_defaults(func=self.validate_config)

        update_local = subparsers.add_parser(
            "update-local", help="Create a commit locally updating all mirrors"
        )
        update_local.set_defaults(func=self.update_local)

        update = subparsers.add_parser(
            "update",
            help=(
                "Create a commit locally updating all mirrors and "
                "push to enabled remote target(s)"
            ),
        )
        add_dryrun(update)
        update.set_defaults(func=self.update)

        for p in (update_local, update):
            p.add_argument(
                "--allow-empty",
                action="store_true",
                default=False,
                help=(
                    "Create an empty update commit even if there are no "
                    "updates; primarily for testing purposes"
                ),
            )
            p.add_argument(
                "--skip",
                action="append",
                default=[],
                help="Skip update of these mirror dirs (comma-separated)",
            )

        promote = subparsers.add_parser(
            "promote",
            help=("Promote formerly merged mirror-tool MRs to an additional branch"),
        )
        add_dryrun(promote)
        promote.set_defaults(func=self.promote)

        # TODO: a command to close outstanding PR if any.

        gitlab_ci_yml = subparsers.add_parser(
            "gitlab-ci-yml",
            help=("Print a recommended GitLab CI configuration to stdout"),
        )
        gitlab_ci_yml.set_defaults(func=self.gitlab_ci_yml)

        return parser

    @property
    def config(self) -> Config:
        if not self._config:
            self._config = Config.from_file(self.args.conf)
        return self._config

    @property
    def skip(self) -> list[str]:
        out: list[str] = []
        for arg in getattr(self.args, "skip", []):
            out.extend(arg.split(","))
        return out

    def run_cmd(
        self, args, check=True, silent=False, env=None, capture_output=None
    ) -> subprocess.CompletedProcess:
        if not silent:
            LOG.info("+ %s" % " ".join(args))
        return subprocess.run(args, check=check, env=env, capture_output=capture_output)

    def run_git_cmd(self, *args, **kwargs):
        kwargs["env"] = environ_with_git_config(self.config.git_config, os.environ)
        return self.run_cmd(*args, **kwargs)

    def commitmsg_for_update(self, update: UpdateInfo) -> str:
        jinja_env = jinja2.Environment(
            loader=jinja2.DictLoader({"commitmsg": self.config.commitmsg})
        )
        return jinja_env.get_template("commitmsg").render(jinja_args(**asdict(update)))

    def update_local_mirror(self, mirror: Mirror) -> UpdateInfo:
        self.run_git_cmd(
            ["git", "fetch", mirror.url, f"+{mirror.ref}:refs/mirror-tool/to-merge"]
        )

        revision = subprocess.check_output(
            ["git", "rev-parse", "refs/mirror-tool/to-merge"], text=True
        ).strip()
        update_info = get_update_info(rev_from="HEAD", rev_to=revision, mirror=mirror)

        self.run_git_cmd(
            [
                "git",
                "merge",
                "-s",
                "ours",
                "--no-commit",
                "--allow-unrelated-histories",
                "refs/mirror-tool/to-merge",
            ]
        )

        if os.path.exists(mirror.dir):
            self.run_git_cmd(
                ["git", "rm", "--quiet", "-rf", f"{mirror.dir}/"], check=False
            )

        self.run_git_cmd(
            [
                "git",
                "read-tree",
                f"--prefix={mirror.dir}/",
                "-u",
                "refs/mirror-tool/to-merge",
            ]
        )

        if update_info.changed or self.args.allow_empty:
            commitmsg = self.commitmsg_for_update(update_info)
            commit_cmd = ["git", "commit", "-m", commitmsg]
            if self.args.allow_empty:
                commit_cmd.append("--allow-empty")

            self.run_git_cmd(commit_cmd)

        return update_info

    def update_local(self) -> list[UpdateInfo]:
        cfg = self.config

        updates = []

        for mirror in cfg.mirrors:
            if mirror.dir in self.skip:
                LOG.info("Skipping update of %s", mirror.dir)
                continue
            updates.append(self.update_local_mirror(mirror))

        LOG.info("Mirror(s) locally updated.")

        return [u for u in updates if u.changed]

    def update(self):
        updates = self.update_local()

        if not self.config.gitlab_merge.enabled:
            LOG.info("No remote targets are enabled for update.")
            return

        gitlab = GitlabUpdateSession(
            self.config.gitlab_merge,
            run_cmd=self.run_cmd,
            updates=updates,
            dry_run=self.args.dry_run,
        )
        gitlab.ensure_merge_request_exists()

    def promote(self):
        if not self.config.gitlab_promote:
            LOG.info("No remote targets have any promotion rules.")
            return

        for promote in self.config.gitlab_promote:
            LOG.info("Checking %s => %s promotion...", promote.src, promote.dest)
            gitlab = GitlabPromoteSession(
                promote, run_cmd=self.run_cmd, dry_run=self.args.dry_run
            )
            gitlab.ensure_promotion_merge_request_exists()

    def gitlab_ci_yml(self):
        if not self.config.gitlab_merge.enabled:
            LOG.info("GitLab features are not enabled in config.")
            return

        print(render_ci_template_from_config(self.config))

    def validate_config(self):
        try:
            self.config.validate()
        except ValidationError as ex:
            LOG.error(
                "%s: configuration error\n  Path: %s\n  Object: %s\n  Cause: %s",
                self.args.conf,
                ".".join([str(p) for p in ex.absolute_path]) or "<top level of config>",
                ex.instance,
                ex.message,
            )
            sys.exit(80)

        LOG.info(
            "%s is a valid configuration file defining %s mirror(s).",
            self.args.conf,
            len(self.config.mirrors),
        )

    def no_command(self):
        LOG.error("Must specify a command (try `--help').")
        sys.exit(72)

    def run(self, args):
        logging.basicConfig(level=logging.WARNING, format="%(message)s")
        LOG.setLevel(logging.INFO)
        self.args = self.parser.parse_args(args)
        self.args.func()


def entrypoint():
    MirrorTool().run(sys.argv[1:])
